#ifndef _CP_SOCKET_H
#define _CP_SOCKET_H

/**
 * @addtogroup cp_socket
 */
/** @{ */
/**
 * @file
 * definitions for socket abstraction api
 */

#include "common.h"

__BEGIN_DECLS

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "hashlist.h"
#include "thread.h"

#include <sys/time.h>
#include <netinet/in.h>

#ifdef USE_SSL
#include <openssl/ssl.h>
#endif

#if defined(sun) || defined(__OpenBSD__)
#include <sys/types.h>
#include <sys/socket.h>
#endif

#define CPSOCKET_DEFAULT_BACKLOG                 50
#define CPSOCKET_DEFAULT_DELAY_SEC                1
#define CPSOCKET_DEFAULT_DELAY_USEC               0 
#define CPSOCKET_THREADPOOL_DEFAULT_SIZE_MIN      5
#define CPSOCKET_THREADPOOL_DEFAULT_SIZE_MAX     50

/** recommended to call before using socket functions */
void cp_socket_init();
/** 
 * call from signal handler to stop sockets in waiting select() and close all 
 * connections
 */
void cp_socket_stop_all();
/** performs cleanup */
void cp_socket_shutdown();

#ifdef USE_SSL
void cp_socket_ssl_init();
void cp_socket_ssl_shutdown();
#endif

typedef enum 
{ 
	CPSOCKET_STRATEGY_CALLBACK,
	CPSOCKET_STRATEGY_THREADPOOL
} cp_socket_strategy;

struct _cp_socket; /* fwd, uh, declaration */

typedef void *(*cp_socket_thread_function)(void *);
typedef int (*cp_socket_callback)(struct _cp_socket *, int fd);
#ifdef USE_SSL
typedef int (*cp_socket_ssl_callback)(struct _cp_socket *, SSL *ssl);
#endif

/**
 * cp_socket is a 'server socket'. create a cp_socket to listen on a port, 
 * specify a callback to implemet your communication protocol (libcprops 
 * provides a basic http implementation you could plug in) and the strategy
 * you want for handling handling multiple connections - use a thread pool
 * to serve each connection on its own thread or a more select()-like approach. 
 * <p>
 * If you create a socket with the CPSOCKET_STRATEGY_THREADPOOL, supply a 
 * thread function to handle communication once a connection has been 
 * accept()-ed. Your cp_thread function should close the connection when done 
 * and check the 'closing' flag, which is set to 1 on shutdown. 
 * <p>
 * If you create a socket with the CPSOCKET_STRATEGY_CALLBACK, supply a callback
 * to process incremental feeds as they come along.
 */
typedef struct _cp_socket
{
	int id;
	int port;                     /* the port this socket listens on */
	int backlog;                  /* max backlog */
	struct timeval delay;         /* max blocking time */
	int fd;                       /* listener */
	cp_hashlist *fds;             /* all available file descriptors */
	cp_thread_pool *tpool;        /* cp_thread pool (if requested) */
	int poolsize_min;             /* min cp_thread count for cp_thread pool */
	int poolsize_max;             /* max cp_thread count for cp_thread pool */
	int fdn;                      /* largest fd - needed for select() */
	cp_socket_strategy strategy;  /* cp_threadpool or callback */	
	union
	{
		cp_socket_thread_function thread_fn;
		cp_socket_callback callback;
#ifdef USE_SSL
		cp_socket_ssl_callback ssl_callback;
#endif
	} action;
	int closing; /* shutdown flag */
	void *owner;
#ifdef USE_SSL
	int use_ssl;
	SSL_CTX *ctx;
#endif
} cp_socket;

inline void cp_socket_set_backlog(cp_socket *socket, int backlog);
inline void cp_socket_set_delay(cp_socket *socket, struct timeval delay);
inline void cp_socket_set_delay_sec(cp_socket *socket, long sec);
inline void cp_socket_set_delay_usec(cp_socket *socket, long usec);
inline void cp_socket_set_poolsize_min(cp_socket *socket, int min);
inline void cp_socket_set_poolsize_max(cp_socket *socket, int max);
inline void cp_socket_set_owner(cp_socket *socket, void *owner);

cp_socket *cp_socket_create(int port, cp_socket_strategy strategy, void *fn);
#ifdef USE_SSL
cp_socket *cp_socket_create_ssl(int port, 
		                        cp_socket_strategy strategy, 
								void *fn,
								char *certificate_file,
								char *key_file,
								int verification_mode);
int cp_socket_ssl_connection_close(cp_socket *sock, SSL *ssl);
#endif
void cp_socket_delete(cp_socket *sock);
int cp_socket_listen(cp_socket *sock);
int cp_socket_select(cp_socket *sock);
int cp_socket_connection_close(cp_socket *sock, int fd);

typedef struct _cp_connection_descriptor
{
	cp_socket *sock;
	struct sockaddr_in *addr;
	int fd;
	void *owner; /* placeholder for client identifier */
#ifdef USE_SSL
	SSL *ssl;
#endif
} cp_connection_descriptor;

cp_connection_descriptor *cp_connection_descriptor_create(cp_socket *sock, 
													struct sockaddr_in *addr, 
													int fd);
#ifdef USE_SSL
cp_connection_descriptor *cp_connection_descriptor_create_ssl(cp_socket *sock, 
													struct sockaddr_in *addr, 
													int fd, 
													SSL *ssl);
#endif
void cp_connection_descriptor_destroy(cp_connection_descriptor *conn_desc);

#define cp_connection_descriptor_get_socket(cd) ((cd)->sock)
#define cp_connection_descriptor_get_addr(cd) ((cd)->addr)
#define cp_connection_descriptor_get_fd(cd) ((cd)->fd)

__END_DECLS

/** @} */

#endif /* _CP_SOCKET_H */
