#ifndef _CP_RTREE_H
#define _CP_RTREE_H

/**
 * @addtogroup cp_trie
 */
/** @{ */

#include "common.h"

__BEGIN_DECLS

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include "hashtable.h"
#include "mtab.h"

#define NODE_STORE(n, i, v) mtab_put((n)->others, *i, (v), strdup(i))
#define NODE_MATCH(n, i) ((n)->others ? mtab_get((n)->others, *i) : NULL)
#define BRANCH_COUNT(node) mtab_count((node)->others)

typedef int (*cp_trie_match_fn)(void *leaf); 

struct _cp_trie;

typedef struct _cp_trie_node 
{ 
	mtab *others; 
	void *leaf; 
} cp_trie_node; 


cp_trie_node *cp_trie_node_new(void *leaf); 
void *cp_trie_node_delete(struct _cp_trie *grp, cp_trie_node *node);
void cp_trie_delete_mapping(struct _cp_trie *grp, mtab_node *map_node);

inline void cp_trie_node_unmap(struct _cp_trie *grp, cp_trie_node **node); 

/**
 * @file
 * cp_trie is a character trie implementation. Tries allow for prefix matching
 * with O(m) = O(1) time (m being the length of the key). Used to store key - 
 * value mappings, tries have certain advantages over hashtables in that worse 
 * case behavior is still O(1) and no hash function is needed. 
 */
typedef struct _cp_trie
{ 
	cp_trie_node *root; 
	int path_count; 

	int mode; 

	cp_copy_fn copy_leaf;
	cp_destructor_fn delete_leaf;

	pthread_rwlock_t *lock; 
} cp_trie; 

/** 
 * create a new cp_trie object with the specified collection mode and
 * leaf management functions
 */
cp_trie *cp_trie_create_trie(int mode, 
		                     cp_copy_fn copy_leaf, 
							 cp_destructor_fn delete_leaf);

/** create a new cp_trie object with the specified collection mode */
inline cp_trie *cp_trie_create(int mode);
/** delete a cp_trie object */
int cp_trie_destroy(cp_trie *grp); 
/** add a mapping to a trie */
int cp_trie_add(cp_trie *grp, char *key, void *leaf); 
/** remove a mapping from a trie */
int cp_trie_remove(cp_trie *grp, char *key, void **leaf); 
/** return the mapping for the longest prefix of the given key */
int cp_trie_prefix_match(cp_trie *grp, char *key, void **leaf);
/** return the mapping for the given key if any */
void *cp_trie_exact_match(cp_trie *grp, char *key);
inline int cp_trie_match_first(void *leaf); 
inline int cp_trie_match_last(void *leaf); 

inline void cp_trie_set_root(cp_trie *grp, void *leaf); 

int cp_trie_lock(cp_trie *grp, int type);
#define cp_trie_rdlock(grp) (cp_trie_lock(grp, COLLECTION_LOCK_READ))
#define cp_trie_wrlock(grp) (cp_trie_lock(grp, COLLECTION_LOCK_WRITE))
int cp_trie_unlock(cp_trie *grp);

__END_DECLS

/** @} */

#endif

