#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <unistd.h>
#include <regex.h>
#include <dirent.h>
#include <errno.h>
#include <sys/time.h>

#include <netdb.h>

#if defined(__OpenBSD__) || defined(__NetBSD__) || defined(__FreeBSD__)
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#endif

#include "util.h"
#include "log.h"
#include "common.h"

#ifdef HAVE_CONFIG_H
#include "config.h"

#ifndef HAVE_STRDUP
char *strdup(char *src)
{
	char *dst = NULL;

    if (src != NULL) // || *src == '\0') 
	{
	    int len = strlen(src);
    	dst = (char *) malloc((len + 1) * sizeof(char));
	    cp_assert(*dst != NULL);

		memcpy(*dst, src, len);
		dst[len] = '\0';
	}

    return dst;
}
#endif /* no strdup */

#ifndef HAVE_STRNDUP
char *strndup(char *src, int maxlen)
{
    int len;
	char *p;
	char *dst = NULL;

	if (maxlen < 0)
	{
		cp_error(CP_INVALID_VALUE, "negative string length requested");
		errno = EINVAL;
		return NULL;
	}

    if (src != NULL) 
	{
		/* null termination not guaranteed - can't use strlen */ 
		for (p = src; *p && p - src < maxlen; p++); 
		len = p - src;
		if (len > maxlen) len = maxlen; 
		dst = (char *) malloc((len + 1) * sizeof(char));
		if (dst == NULL)
		{
			errno = ENOMEM;
			return NULL;
		}

		memcpy(dst, src, len); 
		dst[len] = '\0';
	}

    return dst;
}
#endif /* no strndup */

#endif /* HAVE_CONFIG_H */

char *hex_url_encode(char *hex)
{
	int len = strlen(hex) * 3 / 2 + 1;
	char *res = (char *) malloc(len * sizeof(char));
	char *idx = res;

	while (*hex)
	{
		*idx++ = '%';
		*idx++ = *hex++;
		if (*hex == 0) break;
		*idx++ = *hex++;
	}

	*idx = '\0';

	return res;
}

char *strnchr(char *str, char ch, int len)
{
	char *res = NULL;

	if (len < 0)
	{
		cp_error(CP_INVALID_VALUE, "negative string length requested");
		errno = EINVAL;
		return NULL;
	}

	while (len-- > 0)
	{
		if (*str == ch)
		{
			res = str;
			break;
		}
		if (*str++ == '\0')
			break;
	}

	return res;
}


char *str_trim_cpy(char *dst, char *src)
{
    char *top;
    char *last;
    top = dst;
    while (isspace(*src)) src++;
    last = dst;
    while ((*dst = *src)) {
        dst++;
        if (!isspace(*src)) last = dst;
        src++;
    }

    *last = '\0';
    return top;
}

int parse_boolean(char *value)
{
    return value != NULL && (*value == 't' || *value =='T');
}

long get_timestamp()
{
    return time(NULL);
}

/**
 * Static empty cp_string.
 */
static char *empty_cp_string = "";

inline char *dbfmt(char *str)
{
    return str ? str : empty_cp_string;
}

inline unsigned long get_current_ip()
{
    char hostname[256];
    int rc;

    rc = gethostname(hostname, 256);
    return rc ? 0 : get_host_ip(hostname);
}

unsigned long get_host_ip(char *hostname)
{
    struct addrinfo hints, *res, *res0;
    int rc;
    unsigned long ip = 0;
    unsigned long addr = 0;

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;

    rc = getaddrinfo(hostname, "http", &hints, &res0);
    for (res = res0; res; res = res->ai_next) {
        addr = ((struct sockaddr_in *) res->ai_addr)->sin_addr.s_addr;
        if (ip && (addr & 0xff) == 0x7f) continue;
        ip = addr;
    }

    return ip;
}

inline char *ip_to_string(unsigned long ip, char *buf)
{
    sprintf(buf, "%ld.%ld.%ld.%ld", ip & 0xff, (ip >> 8) & 0xff
           , (ip >> 16) & 0xff, (ip >> 24) & 0xff);
    return buf;
}


char *regex_compilation_error(int rc)
{
	char *msg = NULL;

	switch (rc)
	{
		case REG_BADRPT:
			msg = "Invalid use of repetition operators such as using `*' as the first character."; 
			break;

		case REG_BADBR: 
			msg = "Invalid use of back reference operator.";
			break;

		case REG_EBRACE:
			msg = "Un-matched brace interval operators.";
			break;

		case REG_EBRACK:
			msg = "Un-matched bracket list operators.";
			break;

		case REG_ERANGE:
			msg = "Invalid  use  of  the  range operator, eg. the ending point of the range occurs prior to the starting point.";
			break;

   		case REG_ECTYPE:
			msg = "Unknown character class name.";
			break;

   		case REG_ECOLLATE:
			msg = "Invalid collating element.";
			break;

		case REG_EPAREN:
			msg = "Un-matched parenthesis group operators.";
			break;

		case REG_ESUBREG:
			msg = "Invalid back reference to a subexpression.";
			break;

#ifdef linux
		case REG_EEND:
			msg = "Non specific error.";
			break;
#endif

		case REG_EESCAPE:
			msg = "Trailing backslash.";
			break;

		case REG_BADPAT:
			msg = "Invalid use of pattern operators such as group or list.";
			break;

#ifdef linux
		case REG_ESIZE:
			msg = "Compiled regular expression requires a pattern buffer larger than  64Kb.";
			break;
#endif

		case REG_ESPACE:
			msg = "The regex routines ran out of memory.";
			break;
	}

	return msg;
}

char *stat_error_fmt(int err)
{
	char *fmt = "%s: resolving %s: unknown stat error\n";
	switch (err)
	{
		case ENOENT: 
			fmt = "%s: A component of the path \'%s\' does not exist\n";
			break;

		case ENOTDIR: 
			fmt = "%s: A component of the path \'%s\' is not a directory\n";
			break;

		case ELOOP:
			fmt = "%s: Too many symbolic links encountered while "
				  "traversing the path %s\n";
			break;

		case EFAULT:
			fmt = "%s: Bad address resolving %s\n";
			break;

		case EACCES:
			fmt = "%s: %s: Permission denied.\n";
			break;

		case ENOMEM: 
			fmt = "%s: resolving %s: Out of memory\n";
			break;

		case ENAMETOOLONG:
			fmt = "%s: resolving %s: File name too long.\n";
			break;
	}

	return fmt;
}

/**
 * check if the directory specified by path exists
 */
int checkdir(char *path)
{
	int rc = 0;
	DIR *dir = opendir(path);

	/*  opendir may set:
	 *  EACCES  Permission denied.
     *  EMFILE  Too many file descriptors in use by process.
     *  ENFILE  Too many files are currently open in the system.
     *  ENOENT  Directory does not exist, or name is an empty string.
     *  ENOMEM  Insufficient memory to complete the operation.
     *  ENOTDIR name is not a directory.
	 */
	if (dir == NULL)
	{
		int rc = errno;
		char *err = strerror(rc);
		cp_error(rc, "%s: opendir failed - %s", path, err);
	}
	else
		closedir(dir);

	return rc;
}

/**
 * generates a 16 character id based on the current time - buffer must be at
 * least 17 bytes long. 
 */
void gen_id_str(char *buf)
{
	struct timeval tm;
	gettimeofday(&tm, NULL);
	gen_tm_id_str(buf, &tm);
}

static int id_factor[] = {   2,   3,   5,   7,  11, 
                            13,  17,  19,  23,  29, 
                            31,  37,  41,  43,  47, 
                            53,  59,  61,  67,  71,
                            73,  79,  83,  89,  97, 
                           101, 103, 107, 109, 113,
                           127, 131};

/**
 * generates a 16 character id based on the given timeval - buffer must be at
 * least 17 bytes long
 */
void gen_tm_id_str(char *buf, struct timeval *tm)
{
	int i;
	unsigned long t = 0;
	unsigned long v = 0;
	
	memset(buf, 0, 17);
	t = tm->tv_sec;
	for (i = 0; i < 32; i++)
	{
		v |= (t % id_factor[i]);
		v <<= 1;
		buf[i % 16] = 'a' + (((((t * (i + 1)) % 26) | (buf[i % 16] % 'a')) ^ (unsigned char)(v % 26))) % 26;
	}
	t = tm->tv_usec;
	for (i = 0; i < 32; i++)
	{
		v ^= (t % id_factor[i]);
		v <<= 1;
		buf[i % 16] = 'a' + (((((t * (i + 1)) % 26) | (buf[i % 16] % 'a')) ^ (unsigned char)(v % 26))) % 26;
	}
}

